// http://www.cepstrum.co.jp/
//
// RS-232C loopback test program (with buffering)
//   9600bps, N81 without flow control
//
//   DSK --> cntl0 --> MAX232A --> TXD(RS-232C)
//   DSK <-- stat0 <-- MAX232A <-- RXD(RS-232C)
//
//   tx_ready  : 0=(TX busy), 1=(TX accepts data)
//   rx_ready  : 0=(RX data empty), 1=(RX data exists)
//   rx_full   : 0=(RX buffer not full), 1=(RX buffer full)
//   rx_remain : RX data number in buffer
//
// line in, line out, Fs=48kHz   
                      
#include "c6713dsk.h"             

#define UINT         unsigned int
#define USHORT       unsigned short
#define UCHAR        unsigned char

#define PERIOD       1465     // 225MHz/4/1465=38.4kHz(26us) for 9600bps
#define RX_BUF_LEN   160      // RS-232C RX buffer length

#define ALPHA        0.05
#define DRAMBUF_SIZE 480000                                    

#pragma DATA_SECTION(drambuf, "sdram");      
short   drambuf[DRAMBUF_SIZE];
int     ix_drambuf=0;

volatile int   tx_ready=1;
volatile int   rx_ready=0;
volatile int   rx_full=0;
volatile int   rx_remain=0;

volatile int   _tx_empty=1;
volatile int   _tx_timer;
volatile UCHAR _tx_data;
volatile UCHAR _rx_buf[RX_BUF_LEN];  // RX ring buffer
volatile int   _rx_r=0;              // RX ring buffer read pointer
volatile int   _rx_w=0;              // RX ring buffer write pointer
                                         
float   l_dc=0.0;
float   r_dc=0.0;

//================================================================    
// wait routine with LED flash

void wait(int loop) {
  int i, j;

  for (i=1; i<=8; i=i*2) {
    for (j=0; j<loop; j=j+1) *(char *)CPLD=i; 
  }
}  

//================================================================    
// SPI (McBSP0) transmit routine

void spi_tx(USHORT adrs, USHORT data) {
  USHORT tx_stream;
  
  while ((*(UINT *)McBSP0_SPCR&0x00020000)==0);  // wait until tx ready  
  tx_stream=(adrs&0x0f)<<9;
  tx_stream=tx_stream|(data&0x01ff);
  *(USHORT *)McBSP0_DXR=tx_stream;
}
   
//================================================================    
// McBSP1 receive interrupt routine
           
interrupt void mcbsp1_rx_intr(void) {
  UINT temp;
  short l_in, r_in, l_out, r_out;
  
  temp=*(UINT *)McBSP1_DRR;         // read A/D
  l_in=temp>>16;                    // l_in : left ch. in
  r_in=temp&0xffff;                 // r_in : right ch. in      
  l_dc=(1.0-ALPHA)*l_dc+ALPHA*l_in; // accumulate DC level
  r_dc=(1.0-ALPHA)*r_dc+ALPHA*r_in; // accumulate DC level
  l_in=l_in-l_dc;                   // eliminate DC offset
  r_in=r_in-r_dc;                   // eliminate DC offset

  // something to do here
  l_out=drambuf[ix_drambuf];               // read ring buffer
  drambuf[ix_drambuf]=l_in;                // write ring buffer
  ix_drambuf=(ix_drambuf+1)%DRAMBUF_SIZE;  // update buffer pointer
  r_out=r_in;                              // pass through 

  temp=l_out<<16;                   // l_out : left ch. out
  temp=temp|(r_out&0xffff);         // r_out : right ch. out
  *(UINT *)McBSP1_DXR=temp;         // write D/A
}                     
            
//================================================================  
// digital out (cntl0)

void cntl0(unsigned int x) {
  UCHAR c;
  
  c=*(UCHAR *)(CPLD+1);
  *(UCHAR *)(CPLD+1)=(c&0xfe)|(x&0x01);
}

//================================================================  
// digital in (stat0);

unsigned int stat0(void) {
  return (*(UCHAR *)(CPLD+1)>>4)&0x01;
}

//================================================================    

int tx232c(UCHAR x) {
  if (tx_ready) {
    tx_ready=0;
    _tx_data=x;
    _tx_empty=0;
    _tx_timer=0;
    return 1;
  } else {
    return 0;
  }
}

//================================================================    

UCHAR rx232c() {
  UCHAR x;
  
  if (rx_ready) {
    x=_rx_buf[_rx_r];
    _rx_r=(_rx_r+1)%RX_BUF_LEN;
    if (_rx_r==_rx_w) rx_ready=0;
    rx_full=0;
    rx_remain=rx_remain-1;
    return x;
  } else {
    return 0;
  }
}

//================================================================    

interrupt void timer0_intr(void) {  // INT14
  static UCHAR rx_data;
  static int   rx_timer=0;   
  
  // RS-232C TX processing
  if (!_tx_empty) {
    switch (_tx_timer) {
      case 0:                // start bit
        cntl0(0);
        break;
      case 4:                // b0
        cntl0(_tx_data);
        break;  
      case 8:                // b1
        cntl0(_tx_data>>1);
        break;
      case 12:               // b2
        cntl0(_tx_data>>2);
        break;
      case 16:               // b3
        cntl0(_tx_data>>3);
        break;
      case 20:               // b4
        cntl0(_tx_data>>4);
        break;
       case 24:              // b5
        cntl0(_tx_data>>5);
        break;
      case 28:               // b6
        cntl0(_tx_data>>6);
        break;
      case 32:               // b7
        cntl0(_tx_data>>7);      
        break;
      case 36:               // stop bit
        cntl0(1);
        break;
      case 39:               // transmit end (use value 39 
        tx_ready=1;          // instead of 40 to increase
        _tx_empty=1;         // timing margin of loopback test)
        break;               
      default:
        ;
    }
    _tx_timer=_tx_timer+1;
  }
  
  // RS-232C RX processing
  if (rx_timer==0) {
    if (stat0()==0) rx_timer=1;     // sense start bit
  } else { 
    switch (rx_timer) {
      case 5:                // b0
        rx_data=stat0();
        break;
      case 9:                // b1
        rx_data=rx_data|(stat0()<<1);
        break;
      case 13:               // b2
        rx_data=rx_data|(stat0()<<2);
        break;
      case 17:               // b3
        rx_data=rx_data|(stat0()<<3);
        break;
      case 21:               // b4
        rx_data=rx_data|(stat0()<<4);
        break;             
      case 25:               // b5
        rx_data=rx_data|(stat0()<<5);
        break;
      case 29:               // b6
        rx_data=rx_data|(stat0()<<6);
        break;       
      case 33:               // b7
        rx_data=rx_data|(stat0()<<7);
        if (!rx_full) {
          _rx_buf[_rx_w]=rx_data;
          _rx_w=(_rx_w+1)%RX_BUF_LEN;
          if (_rx_w==_rx_r) rx_full=1;
          rx_ready=1;        
          rx_remain=rx_remain+1;
        }
        break;
      default:
        ;
    }
    rx_timer=(rx_timer+1)%37;
  }

}

//================================================================  

void init_rs232c(unsigned int period) {
  int temp;

  *(UINT*)TIMER0_CTRL=0;            // stop timer0
  *(UINT*)TIMER0_PRD =period;       // set timer0 period
  *(UINT*)TIMER0_CTRL=0x00000201;   // set timer0 paramter
  *(UINT*)TIMER0_CTRL=0x000002c1;   // start timer0

  CSR=0x100;                        // disable all interrupts
  temp=*(UINT *)MUXH&0xfc1fffff;    // read interrupt multipexer (MUXH)
  temp=temp|0x00200000;             // INTSEL14 value (TINT0)
  *(UINT*)MUXH=temp;                // set interrupt multiplexer (MUXH)  
  ICR=0xffff;                       // clear all pending interrupt
  IER=IER|0x4000;                   // enable INT14
  CSR=CSR|1;                        // enable interrupt (global)
  
  // initialize RS-232C tx data line
  for (temp=0; temp<1000; temp=temp+1) {
    cntl0(1);                      
  }
}

//================================================================    

void main(void) {          
  UINT  temp;     
  UCHAR data232c;
  
  CSR                 =0x00000100;  // disable all interrupts
  *(UINT *)CCFG       =0;           // cache off
  *(UINT *)DEVCFG     =0;           // ECLK<--SYSCLK3, McASP disable
 
  // PLL setup 
  *(UINT *)PLLCSR     =0x00000000;  // bypass PLL
  wait(1000);                       // wait
  *(UINT *)PLLCSR     =0x00000008;  // reset PLL
  *(UINT *)PLLDIV0    =0x00008000;  // PLLREF=50MHz/1=50MHz
  *(UINT *)PLLM       =0x00000009;  // PLLOUT=50MHz*9=450MHz
  *(UINT *)OSCDIV1    =0x00000004;  // CLKOUT3=50MHz/5=10MHz
  *(UINT *)PLLDIV3    =0x00008004;  // ECLKOUT=450MHz/5=90MHz
  wait(1000);                       // wait
  *(UINT *)PLLDIV2    =0x00008003;  // SYSCLK2=450MHz/4=112.5MHz
  wait(1000);                       // wait
  *(UINT *)PLLDIV1    =0x00008001;  // SYSCLK1=450MHz/2=225MHz  
  wait(1000);                       // wait
  *(UINT *)PLLCSR     =0x00000000;  // bring PLL out of reset
  wait(1000);                       // wait
  *(UINT *)PLLCSR     =0x00000001;  // enable PLL
 
  // EMIF setting       
  *(UINT *)EMIF_GBLCTL=0x00003078;  // EMIF global control
  *(UINT *)EMIF_CECTL0=0xffffff37;  // EMIF CE0 control, 32bit SDRAM   
  *(UINT *)EMIF_CECTL1=0xffffff07;  // EMIF CE1 control, 8bit async
  *(UINT *)EMIF_CECTL2=0xffffff27;  // EMIF CE2 control, 32bit async
  *(UINT *)EMIF_CECTL3=0xffffff27;  // EMIF CE3 control, 32bit async

  // EMIF SDRAM control setting for MT48LC2M32B2-6 (Micron)
  *(UINT *)EMIF_SDTIM =0x00000578;  // SDRAM refresh period          
  *(UINT *)EMIF_SDEXT =0x000a8529;  // SDRAM extension  
  *(UINT *)EMIF_SDCTL =0x47115000;  // SDRAM control                         

  // daughter card reset  
  *(char *)(CPLD+1)   =0x08;        // daughter card reset on
  wait(10000);                      // wait
  *(char *)(CPLD+1)   =0;           // daughter card reset off

  // McBSP0 setting (CODEC control, DSP is master, SPI)
  *(UINT *)McBSP0_SPCR=0x00000000;  // reset McBSP0
  *(UINT *)McBSP0_SRGR=0x20000020;  // clock=112.5MHz/33=3.4MHz
  *(UINT *)McBSP0_PCR =0x00000a0a;  // use internal clock
  *(UINT *)McBSP0_XCR =0x00000040;  // 16bit/word
  *(UINT *)McBSP0_SPCR=0x00c01000;  // enable sample rate clock
  wait(1000);                       // wait
  *(UINT *)McBSP0_SPCR=0x00c11000;  // enable McBSP0 tx
  wait(1000);                       // wait
  
  // CODEC setting
  spi_tx( 9, 0x000);                // CODEC disable  
  spi_tx( 0, 0x017);                // enable line input (left)
  spi_tx( 1, 0x017);                // enable line input (right)
  spi_tx( 2, 0x0f9);                // enable headphone out (left)
  spi_tx( 3, 0x0f9);                // enable headphone out (right)
  spi_tx( 4, 0x012);                // enable line input
  spi_tx( 5, 0x000);                // enable HPF, disable de-emphasis
  spi_tx( 6, 0x000);                // all power on
  spi_tx( 7, 0x043);                // CODEC is master, DSP mode
  spi_tx( 8, 0x001);                // Fs=48kHz (USB mode, MCLK=12MHz)
  spi_tx( 9, 0x001);                // CODEC enable
    
  // McBSP1 setting (CODEC data, DSP is slave)
  *(UINT *)McBSP1_SPCR=0x00000000;  // reset McBSP1
  *(UINT *)McBSP1_PCR =0x00000000;  // use external clock
  *(UINT *)McBSP1_RCR =0x000100a0;  // 32bit/word, 1bit data delay
  *(UINT *)McBSP1_XCR =0x000000a0;  // 32bit/word
  wait(1000);                       // wait
  *(UINT *)McBSP1_SPCR=0x00010001;  // enable McBSP1   
  
  temp=*(UINT *)MUXL&~(0x1f<<21);   // read interrupt multiplexer (MUXL)
  temp=temp|(0x0f<<21);             // set INTSEL8 <-- RINT1
  *(UINT *)MUXL=temp;               // update MUXL
  ICR=0x0000ffff;                   // clear all pending interrupt
  IER=0x00000102;                   // enable NMI & INT8
  CSR=CSR|1;                        // enable interrupt (global)
  
  init_rs232c(PERIOD);              // initialize timer & RS-232C
  
  // forever loop
  while(1) {      
    if ((RX_BUF_LEN/2)<rx_remain) { // if RX buffer half full
      for (temp=0; temp<(RX_BUF_LEN)/2; temp=temp+1) {
        data232c=rx232c();
        while(!tx_ready);           // wait until TX ready
        tx232c(data232c);
      }
    }
  }           
  
}
       
